#!/usr/bin/perl -w
use strict;
use Compress::Zlib ();

# TODO - try nibbles and bits.
my $bits = 8;

sub init_output {
  [""];
}

# Find the number of bits needed for this number
sub bits_needed {
  (unpack "b32", pack "V", $_[0]) =~ /(.*1)/ ? length $1 : die $_[0];
}

open C, ">C";
open D, ">D";
#my $xxx = 4;
sub output {
  my ($state, $code, $max_code) = @_;
  die "Code $code too big" if $code > 65535;
  # warn $code if $xxx-- > 0;
  my $b = bits_needed($max_code);
  my $crossover = 1 << ($b - 1);
  if ($code < $crossover && $code > ($max_code - $crossover)) {
    print C "$code $b $crossover $max_code save\n";
    --$b;
  } else {
    print C "$code $b $crossover $max_code\n";
  }

  my $new = unpack "b$b", pack "V", $code;
  $state->[1] .= $new;
  my $l = length $state->[1];
  if ($l > 8) {
    # Rip multiples of 8 bits off the front
    $state->[0] .= pack "b*", substr $state->[1], 0, $l & ~7, "";
  }
}
sub final {
  my $state = shift;
  return $state->[0] .= pack "b*", $state->[1];
}

sub init_input {
  ["", $_[0]];
}

sub getbits {
  my ($state, $want) = @_;
  my $l = length $state->[0];
  if ($l < $want) {
    return if !length $state->[1]; # EOF
    my $add = 1 + (($want - $l) >> 3);
    # warn "Getting $add $want $l";
    # Put some more bits on the bit accumulator
    $state->[0] .= unpack "b*", substr $state->[1], 0, $add, "";
  }
  return substr $state->[0], 0, $want, "";
}
sub input {
  my ($state, $max_code) = @_;
  my $b = bits_needed($max_code) - 1;
  my $bits = getbits ($state, $b);
  return unless defined $bits;
  # warn "$state->[0] ", length $state->[0];
  my $code = unpack "V", pack "b32", $bits;
  # warn "$code '$state->[0]'";

  my $crossover = 1 << $b;

  if ($code <= ($max_code - $crossover)) {
    $code += $crossover if getbits ($state, 1);
    ++$b; print D "$code $b $crossover $max_code\n";
  } else {
    ++$b; print D "$code $b $crossover $max_code save\n";
  }
  $code;
}

sub compress {
  my $input = shift;
  my $output = init_output;

  my %codes;
  my $next_code = 1<<$bits;
  # TODO - map these to a frequency table with least likely bytes (or nibbles or
  # bits) first.
  $codes{chr ($next_code - $_)} = $_ foreach (0..($next_code-1));

  my $accumulator;
  my $pos = -length $input;
  while ($pos) {
    $accumulator .= substr($input, $pos++, 1);
    if (exists $codes{$accumulator}) {
      # good
      next;
    }
    # Oops. One too many.
    $codes{$accumulator} = $next_code;
    # Back up
    --$pos; chop $accumulator;
    output($output, $codes{$accumulator}, $next_code);
    $next_code++;
    # Yes, we go round the first bit of the loop again once too many.
    $accumulator = '';
  }
  output($output, $codes{$accumulator}, $next_code);
  return final($output);
}

sub uncompress {
  my $input = init_input @_;
  my $output;

  my %codes;
  my $next_code = 1<<$bits;
  # TODO - map these to a frequency table with least likely bytes (or nibbles or
  # bits) first.
  $codes{$_} = chr ($next_code - $_) foreach (0..($next_code-1));

  my $prev = $output = $codes{input ($input, $next_code)};
  while (1) {
    # Not quire sure why we need the +1
    my $code = input ($input, 1 + $next_code);
    last unless defined $code;
    if (exists $codes{$code}) {
      my $this = $codes{$code};
      $output .= $this;
      $codes{$next_code++} = $prev . substr ($this, 0, 1);
      $prev = $this;
    } else {
      die "$next_code $code" unless $next_code == $code;
      $prev .= substr ($prev, 0, 1);
      $codes{$next_code++} = $prev;
      $output .= $prev;
    }
  }
  $output;
}

sub test {
  my ($name, $input) = @_;
  my $compress = compress $input;
  my $def = Compress::Zlib::compress($input);
  printf "We manage %d, Zlib managed %d\n", length $compress, length $def;
  #print "$compress\n";
  my $output = uncompress $compress;
  printf "Output Was %d now %d\n", length $input, length $output;
  print $output eq $input ? "Pass\n" : "Fail\n";
}

# Test code - compress ourselves.
seek DATA, 0, 0 or die $!;
undef $/;
test ("self", <DATA>);
open FH, $^X or die;
test ($^X,<FH>);

__END__
